import ctypes
import numpy as np
from .util import c_lib

gdn_digital_permutation_c = c_lib.gdn_digital_permutation
gdn_digital_permutation_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

dnb2_integer_to_float_c = c_lib.dnb2_integer_to_float
dnb2_integer_to_float_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

lat_gen_linear_c = c_lib.lat_gen_linear
lat_gen_linear_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

gdn_gen_natural_c = c_lib.gdn_gen_natural
gdn_gen_natural_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

dnb2_undo_interlace_c = c_lib.dnb2_undo_interlace
dnb2_undo_interlace_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

dnb2_linear_matrix_scramble_c = c_lib.dnb2_linear_matrix_scramble
dnb2_linear_matrix_scramble_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

fft_bro_1d_radix2_c = c_lib.fft_bro_1d_radix2
fft_bro_1d_radix2_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

gdn_integer_to_float_c = c_lib.gdn_integer_to_float
gdn_integer_to_float_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

dnb2_gen_gray_c = c_lib.dnb2_gen_gray
dnb2_gen_gray_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

lat_gen_natural_c = c_lib.lat_gen_natural
lat_gen_natural_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

gdn_undo_interlace_c = c_lib.gdn_undo_interlace
gdn_undo_interlace_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

dnb2_interlace_c = c_lib.dnb2_interlace
dnb2_interlace_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

dnb2_gen_natural_c = c_lib.dnb2_gen_natural
dnb2_gen_natural_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

fwht_1d_radix2_c = c_lib.fwht_1d_radix2
fwht_1d_radix2_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

dnb2_digital_shift_c = c_lib.dnb2_digital_shift
dnb2_digital_shift_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

gdn_linear_matrix_scramble_c = c_lib.gdn_linear_matrix_scramble
gdn_linear_matrix_scramble_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

gdn_interlace_c = c_lib.gdn_interlace
gdn_interlace_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

lat_gen_gray_c = c_lib.lat_gen_gray
lat_gen_gray_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

gdn_digital_shift_c = c_lib.gdn_digital_shift
gdn_digital_shift_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

lat_shift_mod_1_c = c_lib.lat_shift_mod_1
lat_shift_mod_1_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]

gdn_gen_natural_same_base_c = c_lib.gdn_gen_natural_same_base
gdn_gen_natural_same_base_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

dnb2_gmat_lsb_to_msb_c = c_lib.dnb2_gmat_lsb_to_msb
dnb2_gmat_lsb_to_msb_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_uint64,flags='C_CONTIGUOUS')
]

ifft_bro_1d_radix2_c = c_lib.ifft_bro_1d_radix2
ifft_bro_1d_radix2_c.argtypes = [
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	ctypes.c_uint64,
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS'),
	np.ctypeslib.ndpointer(ctypes.c_double,flags='C_CONTIGUOUS')
]




import numpy as np
import time
import os
import ctypes 
import glob 
import os

c_lib = ctypes.CDLL(glob.glob(os.path.dirname(os.path.abspath(__file__))+"/c_lib*")[0], mode=ctypes.RTLD_GLOBAL)

from .c_funcs import *

def print_opencl_device_info():
    """ Print OpenCL devices info. Copied from https://github.com/HandsOnOpenCL/Exercises-Solutions/blob/master/Exercises/Exercise01/Python/DeviceInfo.py """
    import pyopencl as cl
    platforms = cl.get_platforms()
    for i,p in enumerate(platforms):
        print("Platform %d -------------------------"%i)
        print("\tName:",p.name)
        print("\tVendor:", p.vendor)
        print("\tVersion:", p.version)
        devices = p.get_devices()
        for j,d in enumerate(devices):
            print("\n\tDevice %d -------------------------"%j)
            print("\t\tName:", d.name)
            print("\t\tVersion:", d.opencl_c_version)
            print("\t\tMax. Compute Units:", d.max_compute_units)
            print("\t\tLocal Memory Size:", d.local_mem_size/1024, "KB")
            print("\t\tGlobal Memory Size:", d.global_mem_size/(1024*1024), "MB")
            print("\t\tMax Alloc Size:", d.max_mem_alloc_size/(1024*1024), "MB")
            print("\t\tMax Work-group Total Size:", d.max_work_group_size)
            dim = d.max_work_item_sizes
            print("\t\tMax Work-group Dims:(", dim[0], " ".join(map(str, dim[1:])), ")")
        print()

bs_plugin_indices = {
    #"lat_gen_gray": [3,5,3,4,5],
    #"dnb2_gen_gray": [3,5,7],
}

def get_qmctoolscl_program_from_context(context, func_name, args_device):
    import pyopencl as cl
    FILEDIR = os.path.dirname(os.path.realpath(__file__))
    with open(FILEDIR+"/cl_kernels/%s.cl"%func_name,"r") as kernel_file:
        kernelsource = kernel_file.read()
    if func_name in bs_plugin_indices:
        insert_ints = tuple([int(args_device[i]) for i in bs_plugin_indices[func_name]])
        kernelsource = kernelsource%insert_ints
    program = cl.Program(context,kernelsource).build()
    return program

def _parse_kwargs_backend_queue_program(kwargs):
    if "backend" in kwargs: 
        kwargs["backend"] = kwargs["backend"].lower()
        assert kwargs["backend"] in ["cl","c"] 
    else: 
        kwargs["backend"] = "c"
    if kwargs["backend"]=="cl":
        try:
            import pyopencl as cl
        except:
            raise ImportError("install pyopencl to access these capabilities in qmctoolscl")
        if "context" not in kwargs:
            platform = cl.get_platforms()[kwargs["platform_id"] if "platform_id" in kwargs else 0]
            device = platform.get_devices()[kwargs["device_id"] if "device_id" in kwargs else 0]
            kwargs["context"] = cl.Context([device])
        if "queue" not in kwargs:
            if "profile" in kwargs and kwargs["profile"]:
                kwargs["queue"] = cl.CommandQueue(kwargs["context"],properties=cl.command_queue_properties.PROFILING_ENABLE)
            else:
                kwargs["queue"] = cl.CommandQueue(kwargs["context"])

def _preprocess_fft_bro_1d_radix2(*args_device,kwargs):
    if kwargs["backend"]=="cl" and (kwargs["local_size"] is None or kwargs["local_size"][2]!=kwargs["global_size"][2]):
        raise Exception("fft_bro_1d_radix2 requires local_size is not None and local_size[2] = %d equals global_size[2] = %d"%(kwargs["local_size"][2],kwargs["global_size"][2]))

def _preprocess_ifft_bro_1d_radix2(*args_device,kwargs):
    if kwargs["backend"]=="cl" and (kwargs["local_size"] is None or kwargs["local_size"][2]!=kwargs["global_size"][2]):
        raise Exception("fft_bro_1d_radix2 requires local_size is not None and local_size[2] = %d equals global_size[2] = %d"%(kwargs["local_size"][2],kwargs["global_size"][2]))

def _preprocess_fwht_1d_radix2(*args_device,kwargs):
    if kwargs["backend"]=="cl" and (kwargs["local_size"] is None or kwargs["local_size"][2]!=kwargs["global_size"][2]):
        raise Exception("fwht_1d_radix2 requires local_size is not None and local_size[2] = %d equals global_size[2] = %d"%(kwargs["local_size"][2],kwargs["global_size"][2]))

def _preprocess_lat_gen_natural(r,n,d,bs_r,bs_n,bs_d,n_start,g,x,kwargs):
    if not ((n_start==0 or np.log2(n_start)%1==0) and ((n+n_start)==0 or np.log2(n+n_start)%1==0)):
        raise Exception("lat_gen_natural requires n_start and n+n_start are either 0 or powers of 2")

_overwrite_args = {
    "fft_bro_1d_radix2": 2, 
    "ifft_bro_1d_radix2": 2, 
}

def parse_gs_bs(gs, args3):
    gs = [min(gs[i],args3[i]) for i in range(3)]
    bs = [np.uint64(np.ceil(args3[i]/gs[i])) for i in range(3)]
    gs = [np.uint64(np.ceil(args3[i]/bs[i])) for i in range(3)]
    return gs,bs

def _opencl_c_func(func):
    func_name = func.__name__
    def wrapped_func(*args, **kwargs):
        _parse_kwargs_backend_queue_program(kwargs)
        args = list(args)
        if kwargs["backend"]=="c":
            t0_perf = time.perf_counter()
            t0_process = time.process_time()
            args = args[:3]+args[:3]+args[3:] # repeat the first 3 args to the batch sizes
            try:
                eval('_preprocess_%s(*args,kwargs=kwargs)'%func_name)
            except NameError: pass
            eval("%s_c(*args)"%func_name)
            tdelta_process = time.process_time()-t0_process 
            tdelta_perf = time.perf_counter()-t0_perf 
            return tdelta_perf,tdelta_process
        else: # kwargs["backend"]=="cl"
            import pyopencl as cl
            t0_perf = time.perf_counter()
            assert "global_size" in kwargs 
            kwargs["global_size"],bs = parse_gs_bs(kwargs["global_size"],args)
            if "local_size" not in kwargs:
                kwargs["local_size"] = None
            args_device = [cl.Buffer(kwargs["context"],cl.mem_flags.READ_WRITE|cl.mem_flags.COPY_HOST_PTR,hostbuf=arg) if isinstance(arg,np.ndarray) else arg for arg in args]
            args_device = args_device[:3]+bs+args_device[3:] # repeat the first 3 args to the batch sizes
            if "program" not in kwargs:
                kwargs["program"] =  get_qmctoolscl_program_from_context(kwargs["context"],func_name,args_device)
            cl_func = getattr(kwargs["program"],func_name)
            try:
                eval('_preprocess_%s(*args_device,kwargs=kwargs)'%func_name)
            except NameError: pass
            t0_process = time.perf_counter()
            event = cl_func(kwargs["queue"],kwargs["global_size"],kwargs["local_size"],*args_device)
            if "wait" not in kwargs or kwargs["wait"]:
                event.wait()
                tdelta_process = time.perf_counter()-t0_process
            else:
                tdelta_process = -1
            if isinstance(args[-1],np.ndarray):
                num_overwrite_args = _overwrite_args[func_name] if func_name in _overwrite_args else 1
                for i in range(-1,-1-num_overwrite_args,-1):
                    cl.enqueue_copy(kwargs["queue"],args[i],args_device[i])
            tdelta_perf = time.perf_counter()-t0_perf
            return tdelta_perf,tdelta_process
    wrapped_func.__doc__ = func.__doc__
    return wrapped_func


import numpy as np 
import time

def random_tbit_uint64s(rng, t, shape):
    """Generate the desired shape of random integers with t bits

Args:
    rng (np.random._generator.Generator): random number generator
    t: (int): number of bits with 0 <= t <= 64
    shape (tuple of ints): shape of resulting integer array"""
    assert 0<=t<=64, "t must be between 0 and 64"
    if t<64: 
        x = rng.integers(0,1<<int(t),shape,dtype=np.uint64)
    else: # t==64
        x = rng.integers(-(1<<63),1<<63,shape,dtype=np.int64)
        negs = x<0
        x[negs] = x[negs]-(-(1<<63))
        x = x.astype(np.uint64)
        x[~negs] = x[~negs]+((1<<63))
    return x

def random_uint64_permutations(rng, n, b):
    """Generate n permutations of 0,...,b-1 into a size (n,b) np.ndarray of np.uint64

Args:
    rng (np.random._generator.Generator): random number generator
    n (int): number of permutations
    b (int): permute 0,...,b-1"""
    x = np.empty((n,b),dtype=np.uint64)
    for i in range(n): 
        x[i] = rng.permutation(b) 
    return x

def lat_get_shifts(rng, r, d):
    """Get random shifts
Args:
    rng (np.random._generator.Generator): random number generator
    r (np.uint64): replications
    d (np.uint64): dimension"""
    shifts = rng.random((r,d))
    return shifts 

def dnb2_get_digital_shifts(rng,r,d,tmax_new):
    """Get random shifts
Args:
    rng (np.random._generator.Generator): random number generator
    r (np.uint64): replications
    d (np.uint64): dimension
    tmax_new (np.uint64): bits in each integer"""
    shifts = random_tbit_uint64s(rng,tmax_new,(r,d))
    return shifts

def dnb2_get_linear_scramble_matrix(rng, r, d, tmax, tmax_new, print_mats):
    """Return a scrambling matrix for linear matrix scrambling

Args:
    rng (np.random._generator.Generator): random number generator
    r (np.uint64): replications
    d (np.uint64): dimension
    tmax (np.uint64): bits in each integer
    tmax_new (np.uint64): bits in each integer of the generating matrix after scrambling
    print_mats (np.uint8): flag to print the resulting matrices"""
    S = np.empty((r,d,tmax_new),dtype=np.uint64) 
    for t in range(tmax_new):
        S[:,:,t] = random_tbit_uint64s(rng,min(t,tmax),(r,d))
    S[:,:,:tmax] <<= np.arange(int(tmax),0,-1,dtype=np.uint64)
    S[:,:,:tmax] += np.uint64(1)<<np.arange(int(tmax)-1,-1,-1,dtype=np.uint64)
    if print_mats:
        print("S with shape (r=%d, d=%d, tmax_new=%d)"%(r,d,tmax_new))
        for l in range(r):
            print("l = %d"%l)
            for j in range(d): 
                print("    j = %d"%j)
                for t in range(tmax_new):
                    b = bin(S[l,j,t])[2:]
                    print("        "+"0"*(tmax-len(b))+b)
    return S

def gdn_get_linear_scramble_matrix(rng, r, d, tmax, tmax_new, r_b, bases):
    """Return a scrambling matrix for linear matrix scrambling

Args:
    rng (np.random._generator.Generator): random number generator
    r (np.uint64): replications
    d (np.uint64): dimension
    tmax (np.uint64): bits in each integer
    tmax_new (np.uint64): bits in each integer of the generating matrix after scrambling
    r_b (np.uint64): replications of bases 
    bases (np.ndarray of np.uint64): bases of size r_b*d"""
    S = np.empty((r,d,tmax_new,tmax),dtype=np.uint64)
    bases_2d = np.atleast_2d(bases)
    lower_flag = np.tri(int(tmax_new),int(tmax),k=-1,dtype=bool)
    n_lower_flags = lower_flag.sum()
    diag_flag = np.eye(tmax_new,tmax,dtype=bool)
    for l in range(r):
        l_b = int(l%r_b)
        for j in range(d):
            b = bases_2d[l_b,j]
            Slj = np.zeros((tmax_new,tmax),dtype=np.uint64)
            Slj[lower_flag] = rng.integers(0,b,n_lower_flags)
            Slj[diag_flag] = rng.integers(1,b,tmax)
            S[l,j] = Slj
    return S

def gdn_get_halton_generating_matrix(r,d,mmax):
    """Return the identity matrices comprising the Halton generating matrices
    
Arg:
    r (np.uint64): replications 
    d (np.uint64): dimension 
    mmax (np.uint64): maximum number rows and columns in each generating matrix"""
    return np.tile(np.eye(mmax,dtype=np.uint64)[None,None,:,:],(int(r),int(d),1,1))

def gdn_get_digital_shifts(rng, r, d, tmax_new, r_b, bases):
    """Return digital shifts for gdn

Args: 
    rng (np.random._generator.Generator): random number generator
    r (np.uint64): replications 
    d (np.uint64): dimension 
    tmax_new (np.uint64): number of bits in each shift 
    r_b (np.uint64): replications of bases 
    bases (np.ndarray of np.uint64): bases of size r_b*d"""
    shifts = np.empty((r,d,tmax_new),dtype=np.uint64)
    bases_2d = np.atleast_2d(bases)
    for l in range(r):
         l_b = int(l%r_b)
         for j in range(d):
             b = bases_2d[l_b,j]
             shifts[l,j] = rng.integers(0,b,tmax_new,dtype=np.uint64)
    return shifts

def gdn_get_digital_permutations(rng, r, d, tmax_new, r_b, bases):
    """Return permutations for gdn

Args: 
    rng (np.random._generator.Generator): random number generator
    r (np.uint64): replications 
    d (np.uint64): dimension 
    tmax_new (np.uint64): number of bits in each shift 
    r_b (np.uint64): replications of bases 
    bases (np.ndarray of np.uint64): bases of size r_b*d"""
    bases_2d = np.atleast_2d(bases)
    bmax = bases_2d.max()
    perms = np.zeros((r,d,tmax_new,bmax),dtype=np.uint64)
    for l in range(r):
        l_b = int(l%r_b)
        for j in range(d):
            b = bases_2d[l_b,j]
            for t in range(tmax_new):
                perms[l,j,t,:b] = rng.permutation(b)
    return perms

class NUSNode_dnb2(object):
    def __init__(self, shift_bits=None, xb=None, left_b2=None, right_b2=None):
        self.shift_bits = shift_bits
        self.xb = xb 
        self.left_b2 = left_b2 
        self.right_b2 = right_b2

def dnb2_nested_uniform_scramble(
    r,
    n, 
    d,
    r_x,
    tmax,
    tmax_new,
    rngs,
    root_nodes,
    xb,
    xrb):
    """Nested uniform scramble of digital net b2

Args: 
    r (np.uint64): replications 
    n (np.uint64): points
    d (np.uint64): dimensions
    r_x (np.uint64): replications of xb
    tmax (np.uint64): maximum number of bits in each integer
    tmax_new (np.uint64): maximum number of bits in each integer after scrambling
    rngs (np.ndarray of numpy.random._generator.Generator): random number generators of size r*d
    root_nodes (np.ndarray of NUSNode_dnb2): root nodes of size r*d
    xb (np.ndarray of np.uint64): array of unrandomized points of size r*n*d
    xrb (np.ndarray of np.uint64): array to store scrambled points of size r*n*d"""
    t0_perf = time.perf_counter()
    t0_process = time.process_time()
    t_delta = np.uint64(tmax_new-tmax)
    for l in range(r):
        l_x = np.uint64(l%r_x)
        for j in range(d):
            rng = rngs[l,j]
            root_node = root_nodes[l,j]
            assert isinstance(root_node,NUSNode_dnb2)
            if root_node.shift_bits is None:
                # initilize root nodes 
                assert root_node.xb is None and root_node.left_b2 is None and root_node.right_b2 is None
                root_node.xb = np.uint64(0) 
                root_node.shift_bits = random_tbit_uint64s(rng,tmax_new,1)[0]
            for i in range(n):
                _xb_new = xb[l_x,i,j]<<t_delta
                _xb = _xb_new
                node = root_nodes[l,j]
                t = int(tmax_new)
                shift = np.uint64(0)                 
                while t>0:
                    b = int(_xb>>np.uint64(t-1))&1 # leading bit of _xb
                    ones_mask_tm1 = np.uint64(2**(t-1)-1)
                    _xb_next = _xb&ones_mask_tm1 # drop the leading bit of _xb 
                    if node.xb is None: # this is not a leaf node, so node.shift_bits in [0,1]
                        if node.shift_bits: shift += np.uint64(2**(t-1)) # add node.shift_bits to the shift
                        if b==0: # looking to move left
                            if node.left_b2 is None: # left node does not exist
                                shift_bits = np.uint64(rng.integers(0,2**(t-1))) # get (t-1) random bits
                                node.left_b2 = NUSNode_dnb2(shift_bits,_xb_next,None,None) # create the left node 
                                shift += shift_bits # add the (t-1) random bits to the shift
                                break
                            else: # left node exists, so move there 
                                node = node.left_b2
                        else: # b==1, looking to move right
                            if node.right_b2 is None: # right node does not exist
                                shift_bits = np.uint64(rng.integers(0,2**(t-1))) # get (t-1) random bits
                                node.right_b2 = NUSNode_dnb2(shift_bits,_xb_next,None,None) # create the right node
                                shift += shift_bits # add the (t-1) random bits to the shift
                                break 
                            else: # right node exists, so move there
                                node = node.right_b2
                    elif node.xb==_xb: # this is a leaf node we have already seen before!
                        shift += node.shift_bits
                        break
                    else: #  node.xb!=_xb, this is a leaf node where the _xb values don't match
                        node_b = int(node.xb>>np.uint64(t-1))&1 # leading bit of node.xb
                        node_xb_next = node.xb&ones_mask_tm1 # drop the leading bit of node.xb
                        node_shift_bits_next = node.shift_bits&ones_mask_tm1 # drop the leading bit of node.shift_bits
                        node_leading_shift_bit = int(node.shift_bits>>np.uint64(t-1))&1
                        if node_leading_shift_bit: shift += np.uint64(2**(t-1))
                        if node_b==0 and b==1: # the node will move its contents left and the _xb will go right
                            node.left_b2 = NUSNode_dnb2(node_shift_bits_next,node_xb_next,None,None)  # create the left node from the current node
                            node.xb,node.shift_bits = None,node_leading_shift_bit # reset the existing node
                            # create the right node 
                            shift_bits = np.uint64(rng.integers(0,2**(t-1))) # (t-1) random bits for the right node
                            node.right_b2 = NUSNode_dnb2(shift_bits,_xb_next,None,None)
                            shift += shift_bits
                            break
                        elif node_b==1 and b==0: # the node will move its contents right and the _xb will go left
                            node.right_b2 = NUSNode_dnb2(node_shift_bits_next,node_xb_next,None,None)  # create the right node from the current node
                            node.xb,node.shift_bits = None,node_leading_shift_bit # reset the existing node
                            # create the left node 
                            shift_bits = np.uint64(rng.integers(0,2**(t-1))) # (t-1) random bits for the left node
                            node.left_b2 = NUSNode_dnb2(shift_bits,_xb_next,None,None)
                            shift += shift_bits
                            break
                        elif node_b==0 and b==0: # move the node contents and _xb to the left
                            node.left_b2 = NUSNode_dnb2(node_shift_bits_next,node_xb_next,None,None) 
                            node.xb,node.shift_bits = None,node_leading_shift_bit # reset the existing node
                            node = node.left_b2
                        elif node_b==1 and b==1: # move the node contents and _xb to the right 
                            node.right_b2 = NUSNode_dnb2(node_shift_bits_next,node_xb_next,None,None) 
                            node.xb,node.shift_bits = None,node_leading_shift_bit # reset the existing node
                            node = node.right_b2
                    t -= 1
                    _xb = _xb_next
                xrb[l,i,j] = _xb_new^shift
    tdelta_process = time.process_time()-t0_process 
    tdelta_perf = time.perf_counter()-t0_perf
    return tdelta_perf,tdelta_process,

class NUSNode_gdn(object):
    def __init__(self, perm=None, xdig=None, children=None):
        self.perm = perm
        self.xdig = xdig 
        self.children = children

def gdn_nested_uniform_scramble(
    r,
    n, 
    d,
    r_x,
    r_b,
    tmax,
    tmax_new,
    rngs,
    root_nodes,
    bases,
    xdig,
    xrdig):
    """Nested uniform scramble of general digital nets

Args: 
    r (np.uint64): replications 
    n (np.uint64): points
    d (np.uint64): dimensions
    r_x (np.uint64): replications of xb
    r_b (np.uint64): replications of bases
    tmax (np.uint64): maximum number digits in each point representation
    tmax_new (np.uint64): maximum number digits in each point representation after scrambling
    rngs (np.ndarray of numpy.random._generator.Generator): random number generators of size r*d
    root_nodes (np.ndarray of NUSNode_gdn): root nodes of size r*d
    bases (np.ndarray of np.uint64): array of bases of size r*d
    xdig (np.ndarray of np.uint64): array of unrandomized points of size r*n*d*tmax
    xrdig (np.ndarray of np.uint64): array to store scrambled points of size r*n*d*tmax_new"""
    t0_perf = time.perf_counter()
    t0_process = time.process_time()
    for l in range(r): 
        l_b = int(l%r_b)
        l_x = int(l%r_x)
        for j in range(d):
            rng = rngs[l,j]
            root_node = root_nodes[l,j]
            b = bases[l_b,j]
            assert isinstance(root_node,NUSNode_gdn)
            if root_node.perm is None:
                # initilize root nodes
                assert root_node.xdig is None and root_node.children is None
                root_node.xdig = np.zeros(tmax_new,dtype=np.uint64) 
                root_node.perm = random_uint64_permutations(rng,tmax_new,b)
                root_node.children = [None]*b
            for i in range(n):
                node = root_nodes[l,j]
                t = 0
                perm = np.zeros(tmax_new,dtype=np.uint64)         
                while t<tmax:
                    _xdig = np.zeros(np.uint64(tmax_new-t),dtype=np.uint64)
                    _xdig[:np.uint64(tmax-t)] = xdig[l_x,i,j,t:]
                    dig = _xdig[0]
                    if node.xdig is None: # this is not a leaf node, so node.perm is a single permutation
                        perm[t] = node.perm[dig] # set the permuted value
                        if node.children[dig] is None: # child in dig position does not exist
                            node_perm = random_uint64_permutations(rng,np.uint64(tmax_new-t-1),b)
                            node.children[dig] = NUSNode_gdn(node_perm,_xdig[1:],[None]*b)
                            perm[(t+1):] = node_perm[np.arange(int(tmax_new)-t-1,dtype=np.uint64),_xdig[1:]] # digits in _xdig[1:] index node_perm rows
                            break
                        else: # child in dig position exists, so move there 
                            node = node.children[dig]
                    elif (node.xdig==_xdig).all(): # this is a leaf node we have already seen before!
                        perm[t:] = node.perm[np.arange(int(tmax_new)-t,dtype=np.uint64),_xdig] # digits in _xdig index node_perm rows
                        break
                    else: # node.xdig!=_xdig, this is a leaf node where the _xdig values don't match
                        node_dig = node.xdig[0]
                        perm[t] = node.perm[0,dig]
                        # move node contenst to the child in the dig position
                        node.children[node_dig] = NUSNode_gdn(node.perm[1:],node.xdig[1:],[None]*b) 
                        node.perm = node.perm[0]
                        node.xdig = None
                        if node_dig==dig: 
                            node = node.children[dig] 
                        else: # create child node in the dig position
                            dig_node_perm = random_uint64_permutations(rng,np.uint64(tmax_new-t-1),b)
                            node.children[dig] = NUSNode_gdn(dig_node_perm,_xdig[1:],[None]*b) # create a new leaf node
                            perm[(t+1):] = dig_node_perm[np.arange(int(tmax_new)-t-1,dtype=np.uint64),_xdig[1:]] # digits in _xdig[1:] index node_perm rows
                            break
                    t += 1
                xrdig[l,i,j] = perm
    tdelta_process = time.process_time()-t0_process 
    tdelta_perf = time.perf_counter()-t0_perf
    return tdelta_perf,tdelta_process


from .util import _opencl_c_func
from .c_funcs import *

@_opencl_c_func
def gdn_digital_permutation():
    """Permutation of digits for a generalized digital net

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    r_x (np.uint64): replications of xdig
    r_b (np.uint64): replications of bases
    tmax (np.uint64): rows of each generating matrix
    tmax_new (np.uint64): rows of each new generating matrix
    bmax (np.uint64): common permutation size, typically the maximum basis
    perms (np.ndarray of np.uint64): permutations of size r*d*tmax_new*bmax
    xdig (np.ndarray of np.uint64): binary digital net points of size r_x*n*d*tmax
    xdig_new (np.ndarray of np.uint64): float digital net points of size r*n*d*tmax_new"""
    pass

@_opencl_c_func
def dnb2_integer_to_float():
    """Convert base 2 binary digital net points to floats

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    tmaxes (np.ndarray of np.uint64): bits in integers of each generating matrix of size r
    xb (np.ndarray of np.uint64): binary digital net points of size r*n*d
    x (np.ndarray of np.double): float digital net points of size r*n*d"""
    pass

@_opencl_c_func
def lat_gen_linear():
    """Lattice points in linear order

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    g (np.ndarray of np.uint64): pointer to generating vector of size r*d
    x (np.ndarray of np.double): pointer to point storage of size r*n*d"""
    pass

@_opencl_c_func
def gdn_gen_natural():
    """Generalized digital net where the base can be different for each dimension e.g. for the Halton sequence

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    r_b (np.uint64): number of replications of bases
    mmax (np.uint64): columns in each generating matrix
    tmax (np.uint64): rows of each generating matrix
    n_start (np.uint64): starting index in sequence
    bases (np.ndarray of np.uint64): bases for each dimension of size r_b*d
    C (np.ndarray of np.uint64): generating matrices of size r*d*mmax*tmax
    xdig (np.ndarray of np.uint64): generalized digital net sequence of digits of size r*n*d*tmax"""
    pass

@_opencl_c_func
def dnb2_undo_interlace():
    """Undo interlacing of generating matrices in base 2

Args:
    r (np.uint64): replications
    d (np.uint64): dimension of resulting generating matrices 
    mmax (np.uint64): columns in generating matrices
    d_alpha (np.uint64): dimension of interlaced generating matrices
    tmax (np.uint64): bits in integers of original generating matrices 
    tmax_alpha (np.uint64): bits in integers of interlaced generating matrices
    alpha (np.uint64): interlacing factor
    C_alpha (np.ndarray of np.uint64): interlaced generating matrices of size r*d_alpha*mmax
    C (np.ndarray of np.uint64): original generating matrices of size r*d*mmax"""
    pass

@_opencl_c_func
def dnb2_linear_matrix_scramble():
    """Linear matrix scrambling for base 2 generating matrices

Args:
    r (np.uint64): replications
    d (np.uint64): dimension
    mmax (np.uint64): columns in each generating matrix 
    r_C (np.uint64): original generating matrices
    tmax_new (np.uint64): bits in the integers of the resulting generating matrices
    S (np.ndarray of np.uint64): scrambling matrices of size r*d*tmax_new
    C (np.ndarray of np.uint64): original generating matrices of size r_C*d*mmax
    C_lms (np.ndarray of np.uint64): resulting generating matrices of size r*d*mmax"""
    pass

@_opencl_c_func
def fft_bro_1d_radix2():
    """Fast Fourier Transform for inputs in bit reversed order.
FFT is done in place along the last dimension where the size is required to be a power of 2.
Follows a decimation-in-time procedure described in https://www.cmlab.csie.ntu.edu.tw/cml/dsp/training/coding/transform/fft.html.

Args:
    d1 (np.uint64): first dimension
    d2 (np.uint64): second dimension
    n_half (np.uint64): half of the last dimension of size n = 2n_half along which FFT is performed
    twiddler (np.ndarray of np.double): size n vector used to store real twiddle factors
    twiddlei (np.ndarray of np.double): size n vector used to store imaginary twiddle factors 
    xr (np.ndarray of np.double): real array of size d1*d2*n on which to perform FFT in place
    xi (np.ndarray of np.double): imaginary array of size d1*d2*n on which to perform FFT in place"""
    pass

@_opencl_c_func
def gdn_integer_to_float():
    """Convert digits of generalized digital net to floats

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    r_b (np.uint64): replications of bases 
    tmax (np.uint64): rows of each generating matrix
    bases (np.ndarray of np.uint64): bases for each dimension of size r_b*d
    xdig (np.ndarray of np.uint64): binary digital net points of size r*n*d*tmax
    x (np.ndarray of np.double): float digital net points of size r*n*d"""
    pass

@_opencl_c_func
def dnb2_gen_gray():
    """Binary representation of digital net in base 2 in Gray code order

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    n_start (np.uint64): starting index in sequence
    mmax (np.uint64): columns in each generating matrix
    C (np.ndarray of np.uint64): generating matrices of size r*d*mmax
    xb (np.ndarray of np.uint64): binary digital net points of size r*n*d"""
    pass

@_opencl_c_func
def lat_gen_natural():
    """Lattice points in natural order

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    n_start (np.uint64): starting index in sequence
    g (np.ndarray of np.uint64): pointer to generating vector of size r*d 
    x (np.ndarray of np.double): pointer to point storage of size r*n*d"""
    pass

@_opencl_c_func
def gdn_undo_interlace():
    """Undo interlacing of generating matrices

Args:
    r (np.uint64): replications
    d (np.uint64): dimension of resulting generating matrices 
    mmax (np.uint64): columns in generating matrices
    d_alpha (np.uint64): dimension of interlaced generating matrices
    tmax (np.uint64): rows of original generating matrices
    tmax_alpha (np.uint64): rows of interlaced generating matrices
    alpha (np.uint64): interlacing factor
    C_alpha (np.ndarray of np.uint64): interlaced generating matrices of size r*d_alpha*mmax*tmax_alpha
    C (np.ndarray of np.uint64): original generating matrices of size r*d*mmax*tmax"""
    pass

@_opencl_c_func
def dnb2_interlace():
    """Interlace generating matrices or transpose of point sets to attain higher order digital nets in base 2

Args:
    r (np.uint64): replications
    d_alpha (np.uint64): dimension of resulting generating matrices 
    mmax (np.uint64): columns of generating matrices
    d (np.uint64): dimension of original generating matrices
    tmax (np.uint64): bits in integers of original generating matrices
    tmax_alpha (np.uint64): bits in integers of resulting generating matrices
    alpha (np.uint64): interlacing factor
    C (np.ndarray of np.uint64): original generating matrices of size r*d*mmax
    C_alpha (np.ndarray of np.uint64): resulting interlaced generating matrices of size r*d_alpha*mmax"""
    pass

@_opencl_c_func
def dnb2_gen_natural():
    """Binary representation of digital net in base 2 in natural order

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    n_start (np.uint64): starting index in sequence
    mmax (np.uint64): columns in each generating matrix
    C (np.ndarray of np.uint64): generating matrices of size r*d*mmax
    xb (np.ndarray of np.uint64): binary digital net points of size r*n*d"""
    pass

@_opencl_c_func
def fwht_1d_radix2():
    """Fast Walsh-Hadamard Transform for real valued inputs.
FWHT is done in place along the last dimension where the size is required to be a power of 2.
Follows the divide-and-conquer algorithm described in https://en.wikipedia.org/wiki/Fast_Walsh%E2%80%93Hadamard_transform

Args:
    d1 (np.uint64): first dimension
    d2 (np.uint64): second dimension
    n_half (np.uint64): half of the last dimension along which FWHT is performed
    x (np.ndarray of np.double): array of size d1*d2*2n_half on which to perform FWHT in place"""
    pass

@_opencl_c_func
def dnb2_digital_shift():
    """Digital shift base 2 digital net

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    r_x (np.uint64): replications of xb
    lshifts (np.ndarray of np.uint64): left shift applied to each element of xb
    xb (np.ndarray of np.uint64): binary base 2 digital net points of size r_x*n*d
    shiftsb (np.ndarray of np.uint64): digital shifts of size r*d
    xrb (np.ndarray of np.uint64): digital shifted digital net points of size r*n*d"""
    pass

@_opencl_c_func
def gdn_linear_matrix_scramble():
    """Linear matrix scramble for generalized digital net

Args:
    r (np.uint64): replications 
    d (np.uint64): dimension 
    mmax (np.uint64): columns in each generating matrix
    r_C (np.uint64): number of replications of C 
    r_b (np.uint64): number of replications of bases
    tmax (np.uint64): number of rows in each generating matrix 
    tmax_new (np.uint64): new number of rows in each generating matrix 
    bases (np.ndarray of np.uint64): bases for each dimension of size r*d 
    S (np.ndarray of np.uint64): scramble matrices of size r*d*tmax_new*tmax
    C (np.ndarray of np.uint64): generating matrices of size r_C*d*mmax*tmax 
    C_lms (np.ndarray of np.uint64): new generating matrices of size r*d*mmax*tmax_new"""
    pass

@_opencl_c_func
def gdn_interlace():
    """Interlace generating matrices or transpose of point sets to attain higher order digital nets

Args:
    r (np.uint64): replications
    d_alpha (np.uint64): dimension of resulting generating matrices 
    mmax (np.uint64): columns of generating matrices
    d (np.uint64): dimension of original generating matrices
    tmax (np.uint64): rows of original generating matrices
    tmax_alpha (np.uint64): rows of interlaced generating matrices
    alpha (np.uint64): interlacing factor
    C (np.ndarray of np.uint64): original generating matrices of size r*d*mmax*tmax
    C_alpha (np.ndarray of np.uint64): resulting interlaced generating matrices of size r*d_alpha*mmax*tmax_alpha"""
    pass

@_opencl_c_func
def lat_gen_gray():
    """Lattice points in Gray code order

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    n_start (np.uint64): starting index in sequence
    g (np.ndarray of np.uint64): pointer to generating vector of size r*d 
    x (np.ndarray of np.double): pointer to point storage of size r*n*d"""
    pass

@_opencl_c_func
def gdn_digital_shift():
    """Digital shift a generalized digital net

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    r_x (np.uint64): replications of xdig
    r_b (np.uint64): replications of bases
    tmax (np.uint64): rows of each generating matrix
    tmax_new (np.uint64): rows of each new generating matrix
    bases (np.ndarray of np.uint64): bases for each dimension of size r_b*d
    shifts (np.ndarray of np.uint64): digital shifts of size r*d*tmax_new
    xdig (np.ndarray of np.uint64): binary digital net points of size r_x*n*d*tmax
    xdig_new (np.ndarray of np.uint64): float digital net points of size r*n*d*tmax_new"""
    pass

@_opencl_c_func
def lat_shift_mod_1():
    """Shift mod 1 for lattice points

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    r_x (np.uint64): replications in x
    x (np.ndarray of np.double): lattice points of size r_x*n*d
    shifts (np.ndarray of np.double): shifts of size r*d
    xr (np.ndarray of np.double): pointer to point storage of size r*n*d"""
    pass

@_opencl_c_func
def gdn_gen_natural_same_base():
    """Generalized digital net with the same base for each dimension e.g. a digital net in base greater than 2

Args:
    r (np.uint64): replications
    n (np.uint64): points
    d (np.uint64): dimension
    mmax (np.uint64): columns in each generating matrix
    tmax (np.uint64): rows of each generating matrix
    n_start (np.uint64): starting index in sequence
    b (np.uint64): common base
    C (np.ndarray of np.uint64): generating matrices of size r*d*mmax*tmax
    xdig (np.ndarray of np.uint64): generalized digital net sequence of digits of size r*n*d*tmax"""
    pass

@_opencl_c_func
def dnb2_gmat_lsb_to_msb():
    """Convert base 2 generating matrices with integers stored in Least Significant Bit order to Most Significant Bit order

Args:
    r (np.uint64): replications
    d (np.uint64): dimension
    mmax (np.uint64): columns in each generating matrix 
    tmaxes (np.ndarray of np.uint64): length r vector of bits in each integer of the resulting MSB generating matrices
    C_lsb (np.ndarray of np.uint64): original generating matrices of size r*d*mmax
    C_msb (np.ndarray of np.uint64): new generating matrices of size r*d*mmax"""
    pass

@_opencl_c_func
def ifft_bro_1d_radix2():
    """Inverse Fast Fourier Transform with outputs in bit reversed order.
FFT is done in place along the last dimension where the size is required to be a power of 2.
Follows a procedure described in https://www.expertsmind.com/learning/inverse-dft-using-the-fft-algorithm-assignment-help-7342873886.aspx.

Args:
    d1 (np.uint64): first dimension
    d2 (np.uint64): second dimension
    n_half (np.uint64): half of the last dimension of size n = 2n_half along which FFT is performed
    twiddler (np.ndarray of np.double): size n vector used to store real twiddle factors
    twiddlei (np.ndarray of np.double): size n vector used to store imaginary twiddle factors 
    xr (np.ndarray of np.double): real array of size d1*d2*n on which to perform FFT in place
    xi (np.ndarray of np.double): imaginary array of size d1*d2*n on which to perform FFT in place"""
    pass




import setuptools
from setuptools import Extension
import os 
import re 

cl_files = [file[:-3] for file in os.listdir("./qmctoolscl/cl_kernels/") if file[-3:]==".cl"]

# for cl_file in cl_files:
#     with open("./qmctoolscl/cl_kernels/%s.cl"%cl_file,"r") as f:
#         cl_content = f.read()
#     c_content = '#include "qmctoolscl.h"\n\n'+cl_content 
#     c_content = c_content.replace("__kernel void","EXPORT void")
#     c_content = c_content.replace("__global ","")
#     c_content = c_content.replace("ulong","unsigned long long")
#     c_content = c_content.replace("get_global_id(0)","0")
#     c_content = c_content.replace("get_global_id(1)","0")
#     c_content = c_content.replace("get_global_id(2)","0")
#     c_content = c_content.replace("barrier(CLK_LOCAL_MEM_FENCE);","")
#     c_content = c_content.replace("barrier(CLK_GLOBAL_MEM_FENCE);","")
#     c_content = c_content.replace("barrier(CLK_LOCAL_MEM_FENCE | CLK_GLOBAL_MEM_FENCE);","")
#     with open("./qmctoolscl/c_funcs/%s.c"%cl_file,"w") as f:
#         f.write(c_content)

setuptools.setup(
    name = "qmctoolscl",
    version = "1.1.3",
    author="Aleksei G Sorokin",
    author_email="asorokin@hawk.iit.edu",
    install_requires = [
        'numpy >= 1.17.0',
    ],
    python_requires = ">=3.5",
    description = "Quasi-Monte Carlo Tools in PyOpenCL and C",
    long_description="Python interface to QMC tools with C and OpenCL backends. See https://qmcsoftware.github.io/QMCToolsCL/",
    long_description_content_type="text/markdown",
    url="https://qmcsoftware.github.io/QMCToolsCL/",
    include_package_data=True,
    packages = [
        'qmctoolscl',
    ],
    ext_modules = [
        Extension(
            name = 'qmctoolscl.c_lib',
            sources = ["./qmctoolscl/c_funcs/python_compat.c"]+["./qmctoolscl/c_funcs/%s.c"%cl_file for cl_file in cl_files]
        )
    ],
)

c_to_ctypes_map = {
    "ulong": "uint64",
    "double": "double",
    "char": "uint8",
}

THISDIR = os.path.dirname(os.path.realpath(__file__))

str_c = "import ctypes\nimport numpy as np\nfrom .util import c_lib\n\n"
str_wf = "from .util import _opencl_c_func\nfrom .c_funcs import *\n\n"
str_init = "from .rand_funcs import *\nfrom .wrapped_funcs import ("
for cl_file in cl_files:
    with open("%s/qmctoolscl/cl_kernels/%s.cl"%(THISDIR,cl_file),"r") as f:
        code = f.read() 
    blocks = re.findall(r'(?<=void\s).*?(?=\s?\))',code,re.DOTALL)
    for block in blocks:
        lines = block.replace("(","").splitlines()
        name = lines[0]
        desc = [] 
        si = 1
        while lines[si].strip()[:2]=="//":
            desc += [lines[si].split("// ")[1].strip()]
            si += 1
        desc = "\n".join(desc)
        args = []
        doc_args = []
        for i in range(si,len(lines)):
            var_input,var_desc = lines[i].split(" // ")
            var_type,var = var_input.replace(",","").split(" ")[-2:]
            if var_type not in c_to_ctypes_map:
                    raise Exception("var_type %s not found in map"%var_type)
            c_var_type = c_to_ctypes_map[var_type]
            if var[0]!="*":
                doc_args += ["%s (np.%s): %s"%(var,c_var_type,var_desc)]
                args += ["ctypes.c_%s"%c_var_type]
            else:
                doc_args += ["%s (np.ndarray of np.%s): %s"%(var[1:],c_var_type,var_desc)]
                args += ["np.ctypeslib.ndpointer(ctypes.c_%s,flags='C_CONTIGUOUS')"%c_var_type]
        doc_args = doc_args[:3]+doc_args[6:] # skip batch size args
        str_c += "%s_c = c_lib.%s\n"%(name,name)
        str_c += "%s_c.argtypes = [\n\t%s\n]\n\n"%(name,',\n\t'.join(args))
        str_wf += '@_opencl_c_func\ndef %s():\n    """%s\n\nArgs:\n    %s"""\n    pass\n\n'%(name,desc.strip(),"\n    ".join(doc_args))
        str_init += "\n\t%s,"%name

with open("%s/qmctoolscl/c_funcs.py"%THISDIR,"w") as f: f.write(str_c)
with open("%s/qmctoolscl/wrapped_funcs.py"%THISDIR,"w") as f: f.write(str_wf)
with open("%s/qmctoolscl/__init__.py"%THISDIR,"w") as f: f.write(str_init+"\n)")

# str_tex = str_tex.replace("np.double","floats")
# str_tex = str_tex.replace("np.ndarray","array")
# str_tex = str_tex.replace("of np.uint64","of ints")
# str_tex = str_tex.replace("np.uint64","ints")
# with open("api.tex","w") as f: f.write(str_tex+"\\end{itemize}")





from .rand_funcs import *
from .wrapped_funcs import (
	gdn_digital_permutation,
	dnb2_integer_to_float,
	lat_gen_linear,
	gdn_gen_natural,
	dnb2_undo_interlace,
	dnb2_linear_matrix_scramble,
	fft_bro_1d_radix2,
	gdn_integer_to_float,
	dnb2_gen_gray,
	lat_gen_natural,
	gdn_undo_interlace,
	dnb2_interlace,
	dnb2_gen_natural,
	fwht_1d_radix2,
	dnb2_digital_shift,
	gdn_linear_matrix_scramble,
	gdn_interlace,
	lat_gen_gray,
	gdn_digital_shift,
	lat_shift_mod_1,
	gdn_gen_natural_same_base,
	dnb2_gmat_lsb_to_msb,
	ifft_bro_1d_radix2,
)